import json
from typing import Any, Dict, Optional, Union

registry = {}
default_encoding_name = None  # type: Optional[str]


def register_encoding(name, *, default=False):
    def decorator(cls):
        global default_encoding_name
        registry[name] = cls
        if default:
            default_encoding_name = name
        return cls

    return decorator


class BaseEncoding:
    def encode(self, value: Any, failsafe: bool = False) -> Union[str, bytes]:
        """
        Encode a value to a string or bytes.

        :param failsafe: When set, hint that the encoder should try hard not to fail,
                         even if it requires loss of fidelity.
        """
        raise NotImplementedError("Encoding not implemented")

    def decode(self, value: Union[str, bytes]) -> Any:
        raise NotImplementedError("Decoding not implemented")


@register_encoding("json", default=True)
class JSONEncoding(BaseEncoding):
    """
    Default (JSON) encoding for kwargs and results.
    """

    # These can be effortlessly overridden in subclasses
    dump_kwargs = {
        "ensure_ascii": False,
        "separators": (",", ":"),
    }
    load_kwargs = {}  # type: ignore
    failsafe_default = str

    def encode(self, value: Any, failsafe: bool = False) -> Union[str, bytes]:
        kwargs = self.dump_kwargs.copy()  # type: Dict[str, Any]
        if failsafe:
            kwargs["default"] = self.failsafe_default
        return json.dumps(
            value,
            **kwargs,
        )

    def decode(self, value: Union[str, bytes]) -> Any:
        if isinstance(value, bytes):
            value = value.decode()
        return json.loads(value, **self.load_kwargs)



class NoSuchJob(Exception):
    pass


class DuplicateJob(Exception):
    pass


class InvalidStatus(Exception):
    pass


class AlreadyAcquired(InvalidStatus):
    pass


class AlreadyResulted(InvalidStatus):
    pass


class InvalidJob(ValueError):
    pass



import logging
from os import getpid
from platform import node
from typing import Iterable, List, Optional, Union

from redis import Redis

from minique.compat import sentry_sdk
from minique.enums import JobStatus
from minique.models.job import Job
from minique.models.queue import Queue
from minique.work.job_runner import JobRunner


class Worker:
    queue_timeout = 1
    job_runner_class = JobRunner

    # This property may be ignored by subclasses, but it's here for convenience's sake.
    allowed_callable_patterns = frozenset()  # type: Iterable[str]

    def __init__(self, redis: Redis, queues: List[Queue]) -> None:
        self.id = self.compute_id()
        self.redis = redis
        self.queues = list(queues)
        self.log = logging.getLogger(
            "{}.{}".format(__name__, self.id.replace(".", "_"))
        )
        assert all(isinstance(q, Queue) for q in queues)

    @classmethod
    def for_queue_names(
        cls, redis: Redis, queue_names: Union[List[str], str], **kwargs
    ) -> "Worker":
        if isinstance(queue_names, str):
            queue_names = [queue_names]
        kwargs.setdefault("queues", [Queue(redis, name) for name in queue_names])
        return cls(redis, **kwargs)

    def compute_id(self) -> str:
        # Override me in a subclass if you like!
        return "w-{node}-{pid}".format(node=node().split(".")[0], pid=getpid())

    def get_next_job(self) -> Optional[Job]:
        rv = self.redis.blpop([q.redis_key for q in self.queues], self.queue_timeout)
        if rv:  # The rv is a 2-tuple (queue name, value)
            job_id = rv[1].decode()
            return Job(self.redis, job_id)
        return None

    def tick(self) -> Optional[Job]:
        job = self.get_next_job()
        if not job:
            return None
        job.ensure_exists()
        if job.status == JobStatus.CANCELLED:  # Simply skip running cancelled jobs
            return None
        runner = self.job_runner_class(worker=self, job=job)
        runner.run()
        return job

    def loop(self) -> None:  # pragma: no cover
        while True:
            try:
                self.tick()
            except KeyboardInterrupt:
                break
            except Exception:  # noqa
                self.log.error("Unexpected worker tick error", exc_info=True)

    def process_exception(
        self,
        excinfo: Optional[tuple] = None,
        context: Optional[dict] = None,
    ) -> None:  # pragma: no cover
        """
        A hook to log or process exceptions.

        If sentry_sdk is installed, this implementation will automatically be use it.

        :param excinfo: Optionally, the sys.exc_info() 3-tuple
        """
        if sentry_sdk:
            with sentry_sdk.push_scope() as scope:
                if context:
                    scope.set_context("minique", context)
                sentry_sdk.capture_exception(excinfo)  # type: ignore



from typing import Iterable, Optional

from redis import Redis


def read_list(
    redis_conn: Redis, key: str, *, chunk_size: int = 4096, last_n: Optional[int] = None
) -> Iterable[bytes]:
    """
    Read a possibly large Redis list in chunks.

    Avoids OOMs on the Redis side.

    :param redis_conn: Redis connection
    :param key: Key
    :param chunk_size: How many lines to read per request.
    :param last_n: Attempt to only read the last N lines.
    :return:
    """
    if not redis_conn.exists(key):
        return

    if chunk_size <= 0:
        chunk_size = 4096

    if last_n and last_n > 0:
        offset = max(0, redis_conn.llen(key) - last_n)
    else:
        offset = 0

    while offset < redis_conn.llen(key):
        # Regarding that - 1 there, see this from https://redis.io/commands/lrange:
        # > Note that if you have a list of numbers from 0 to 100, LRANGE list 0 10
        # > will return 11 elements, that is, the rightmost item is included.
        chunk = redis_conn.lrange(key, offset, offset + chunk_size - 1) or []
        if not chunk:
            break
        yield from chunk
        offset += chunk_size



from typing import TYPE_CHECKING, Optional, Tuple

from redis import Redis

from minique.consts import QUEUE_KEY_PREFIX
from minique.excs import NoSuchJob
from minique.utils import cached_property
from minique.utils.redis_list import read_list

if TYPE_CHECKING:
    from minique.models.job import Job


class Queue:
    def __init__(self, redis: Redis, name: str) -> None:
        self.redis = redis
        self.name = str(name)

    @cached_property
    def redis_key(self):
        return "%s%s" % (QUEUE_KEY_PREFIX, self.name)

    @property
    def length(self) -> int:
        return self.redis.llen(self.redis_key)

    def clear(self):
        """
        Entirely clear this queue. Do not call this unless you're willing to risk orphaned jobs.
        """
        return self.redis.delete(self.redis_key)

    def enqueue_initial(self, job: "Job", payload: dict) -> None:
        assert payload["queue"] == self.name
        with self.redis.pipeline() as p:
            p.hmset(job.redis_key, payload)
            if payload["job_ttl"] > 0:
                p.expire(job.redis_key, payload["job_ttl"])
            p.rpush(self.redis_key, job.id)
            p.execute()

    def get_queue_index(self, job: "Job") -> Optional[int]:
        job_id_bytes = str(job.id).encode()
        for index, value in enumerate(read_list(self.redis, self.redis_key)):
            if value == job_id_bytes:
                return index
        return None

    def ensure_enqueued(self, job: "Job") -> Tuple[bool, int]:
        """
        Ensure the job is in the queue.

        This should only be called if the job itself is in a state that would allow its execution,
        and even then only in dire straits, as it is a fixer.

        :return: A 2-tuple of a boolean and an integer describing the queue position;
                 the boolean is true if the job was indeed re-queued
        """
        if not self.redis.exists(job.redis_key):
            raise NoSuchJob(
                "Job object for {id} has disappeared, can not re-enqueue".format(id=id)
            )

        # NB: This is mildly racy; two processes could be doing this scan concurrently.
        #     Technically having an entry in the queue multiple times shouldn't be a problem, since
        #     the second attempt at acquiring the same job will fail anyway (HSETNX in acquire()),
        #     but some heuristics regarding queue length could be off.

        index = self.get_queue_index(job)
        if index is None:
            # RPUSH: Integer reply: the length of the list after the push operation.
            index = self.redis.rpush(self.redis_key, job.id) - 1
            return (True, index)

        return (False, index)



import fnmatch
import json
import logging
import sys
import time
import traceback
from typing import Any, Callable, Union

from redis import Redis

from minique.compat import TYPE_CHECKING
from minique.enums import JobStatus
from minique.excs import AlreadyAcquired, AlreadyResulted, InvalidJob
from minique.models.job import Job
from minique.utils import _set_current_job, import_by_string

if TYPE_CHECKING:
    from minique.work.worker import Worker


class JobRunner:
    def __init__(self, worker: "Worker", job: Job) -> None:
        self.worker = worker
        self.job = job
        self.redis = job.redis
        assert isinstance(self.redis, Redis)
        self.log = logging.getLogger(
            "{}.{}".format(__name__, str(self.job.id).replace(".", "_"))
        )
        job.ensure_exists()

    def acquire(self) -> None:
        new_acquisition_info = json.dumps(self.get_acquisition_info(), default=str)
        if not self.redis.hsetnx(self.job.redis_key, "acquired", new_acquisition_info):
            raise AlreadyAcquired(
                "job {id} already acquired: {info}".format(
                    id=self.job.id,
                    info=self.job.acquisition_info,
                )
            )
        self.redis.hset(self.job.redis_key, "status", JobStatus.ACQUIRED.value)
        self.redis.persist(self.job.redis_key)

    def get_acquisition_info(self) -> dict:
        # Override me in a subclass if you like!
        return {"worker": self.worker.id, "time": time.time()}

    def execute(self) -> Any:
        func = self.get_callable()
        kwargs = self.job.kwargs
        self.log.info("calling %s(%r)", func, kwargs)
        with _set_current_job(job=self.job):
            return func(**kwargs)

    def verify_callable_name(self, name: str) -> bool:
        if not any(
            fnmatch.fnmatch(name, pat) for pat in self.worker.allowed_callable_patterns
        ):
            raise InvalidJob(
                "Name {} doesn't match any pattern in {}".format(
                    name,
                    self.worker.allowed_callable_patterns,
                )
            )
        return True

    def get_callable(self) -> Callable:
        name = self.job.callable_name
        if not self.verify_callable_name(name):
            raise InvalidJob("Invalid job definition")
        return import_by_string(name)  # type: ignore

    def complete(
        self, success: bool, value: Union[str, bytes], duration: float
    ) -> None:
        assert isinstance(success, bool)
        update_payload = {
            "status": (JobStatus.SUCCESS if success else JobStatus.FAILED).value,
            "duration": float(duration),
        }
        if not self.redis.setnx(self.job.result_redis_key, value):  # pragma: no cover
            raise AlreadyResulted("job {id} already has result".format(id=self.job.id))
        self.redis.hmset(self.job.redis_key, update_payload)  # type: ignore
        # Update expiries to the result TTL for both the job and the result
        self.redis.expire(self.job.result_redis_key, self.job.result_ttl)
        self.redis.expire(self.job.redis_key, self.job.result_ttl)
        if success:
            self.log.info("finished in %f seconds", duration)
        else:
            self.log.warning("errored in %f seconds", duration)

    def run(self) -> None:
        try:
            encoding = self.job.get_encoding()
            self.acquire()
        except Exception as exc:
            self.process_exception(sys.exc_info())
            raise exc  # could have had an exception in process_exception
        interrupt = False
        success = False
        encoded_value = encoding.encode({"error": "unknown"})
        start_time = time.time()
        try:
            value = self.execute()
            encoded_value = encoding.encode(value)
            success = True
        except BaseException as exc:
            success = False
            exc_type, exc_value, exc_tb = excinfo = sys.exc_info()
            error_value = {
                "exception_type": getattr(exc_type, "__qualname__", None),
                "exception_value": str(exc_value),
                "traceback": traceback.format_exc(),
            }
            encoded_value = encoding.encode(
                error_value,
                failsafe=True,
            )
            interrupt = isinstance(exc, KeyboardInterrupt)
            self.process_exception(excinfo)
        finally:
            end_time = time.time()
            self.complete(
                success=success, value=encoded_value, duration=(end_time - start_time)
            )
        if interrupt:  # pragma: no cover
            raise KeyboardInterrupt("Interrupt")

    def process_exception(self, excinfo):
        try:
            self.worker.process_exception(
                excinfo,
                context={
                    "id": str(self.job.id),
                    "queue": str(self.job.queue_name),
                },
            )
        except Exception:  # noqa
            self.log.warning("error running process_exception()", exc_info=True)



import time
from typing import Any, Callable, Dict, Optional, Union

from redis import Redis

import minique.encoding as encoding
from minique.enums import JobStatus
from minique.excs import DuplicateJob
from minique.models.job import Job
from minique.models.queue import Queue
from minique.utils import get_random_pronounceable_string


def enqueue(
    redis: Redis,
    queue_name: str,
    callable: Union[Callable, str],
    kwargs: Optional[Dict[str, Any]] = None,
    job_id: Optional[str] = None,
    job_ttl: int = 0,
    result_ttl: int = 86400 * 7,
    encoding_name: Optional[str] = None,
) -> Job:
    if not encoding_name:
        encoding_name = encoding.default_encoding_name
    encoder = encoding.registry[encoding_name]()
    if not isinstance(callable, str):
        callable = "{module}.{qualname}".format(
            module=callable.__module__,
            qualname=callable.__qualname__,
        )

    if not job_id:
        job_id = get_random_pronounceable_string()
    job = Job(redis, job_id)
    if job.exists:
        raise DuplicateJob("duplicate job: {id}".format(id=job_id))
    payload = {
        "queue": queue_name,
        "callable": str(callable),
        "kwargs": encoder.encode(kwargs or {}),
        "encoding_name": encoding_name,
        "status": JobStatus.NONE.value,
        "ctime": time.time(),
        "job_ttl": int(job_ttl),
        "result_ttl": int(result_ttl),
    }
    queue = Queue(redis, name=queue_name)
    queue.enqueue_initial(job=job, payload=payload)
    job.ensure_exists()
    return job


def get_job(redis: Redis, job_id: str) -> Job:
    job = Job(redis, job_id)
    job.ensure_exists()
    return job


def cancel_job(redis: Redis, job_id: str) -> bool:
    """
    Cancel the job with the given job ID.

    If a worker is already busy with the job, it may not immediately quit,
    and as such, the job is not set to cancelled state.

    :type redis: redis.Redis
    :param job_id: Job ID.
    :raises minique.excs.NoSuchJob: if the job does not exist.
    """
    job = get_job(redis, job_id)
    if not (job.has_finished or job.has_started):
        # Cancel the job and remove it from the queue it may be in.
        redis.hset(job.redis_key, "status", JobStatus.CANCELLED.value)
        redis.lrem(job.get_queue().redis_key, 0, job.id)
        return True
    return False



from typing import Any, Callable, Optional

from minique.models.job import Job
from minique.utils import _set_current_job, import_by_string
from minique.work.job_runner import JobRunner
from minique.work.worker import Worker


class TestJobRunner(JobRunner):
    def execute(self) -> Any:
        func = self.job.replacement_callable or import_by_string(self.job.callable_name)
        if self.job.replacement_kwargs is not None:
            kwargs = self.job.replacement_kwargs
        else:
            kwargs = self.job.kwargs
        with _set_current_job(job=self.job):
            return func(**kwargs)


def run_synchronously(
    job: Job,
    replacement_callable: Optional[Callable] = None,
    replacement_kwargs: Optional[dict] = None,
) -> None:
    job.replacement_callable = replacement_callable
    job.replacement_kwargs = replacement_kwargs
    worker = Worker(job.redis, [])
    TestJobRunner(worker=worker, job=job).run()



import argparse
import logging
from typing import List, Optional

from redis import StrictRedis

from minique.compat import sentry_sdk
from minique.work.worker import Worker


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("-u", "--redis-url", required=True)
    parser.add_argument("-q", "--queues", nargs="+", required=True)
    parser.add_argument("--allow-callable", nargs="+", required=True)
    parser.add_argument("--single-tick", action="store_true")
    return parser


def main(argv: Optional[List[str]] = None) -> None:
    parser = get_parser()
    args = parser.parse_args(argv)
    logging.basicConfig(datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    redis = StrictRedis.from_url(args.redis_url)
    worker = Worker.for_queue_names(redis=redis, queue_names=args.queues)
    worker.allowed_callable_patterns = set(args.allow_callable)
    worker.log.info("Worker initialized")
    if sentry_sdk:
        try:
            sentry_sdk.init()
        except Exception as exc:
            worker.log.warning("Failed to initialize Sentry: %s", exc, exc_info=True)
        else:
            hub = sentry_sdk.hub.Hub.current
            if hub and hub.client and hub.client.dsn:
                worker.log.info("Sentry configured with a valid DSN")

    if args.single_tick:
        worker.tick()
    else:
        worker.loop()



import os
import re

import setuptools


def get_version(package):
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("__version__ = ['\"](.+?)['\"]", init_py).group(1)


if __name__ == "__main__":
    setuptools.setup(
        name="minique",
        description="Minimal Redis job runner",
        version=get_version("minique"),
        url="https://github.com/valohai/minique",
        author="Valohai",
        author_email="hait@valohai.com",
        maintainer="Aarni Koskela",
        maintainer_email="akx@iki.fi",
        license="MIT",
        install_requires=["redis>=2.10.0"],
        extras_require={
            "sentry": [
                "sentry_sdk~=0.19.0",
            ]
        },
        packages=setuptools.find_packages(".", exclude=("minique_tests",)),
        include_package_data=True,
        entry_points={
            "console_scripts": [
                "minique = minique.cli:main",
            ],
        },
    )



import json
import time
from typing import Any, Callable, Optional, Tuple

from redis import Redis

from minique import encoding
from minique.compat import TYPE_CHECKING
from minique.consts import JOB_KEY_PREFIX, RESULT_KEY_PREFIX
from minique.enums import JobStatus
from minique.excs import AlreadyAcquired, AlreadyResulted, InvalidStatus, NoSuchJob

if TYPE_CHECKING:
    from minique.models.queue import Queue


class Job:
    # Used for testing:
    replacement_callable = None  # type: Optional[Callable]
    replacement_kwargs = None  # type: Optional[dict]

    def __init__(self, redis: Redis, id: str) -> None:
        self.redis = redis
        self.id = str(id)

    def ensure_exists(self) -> None:
        if not self.exists:
            raise NoSuchJob("Job {id} does not exist".format(id=self.id))

    def ensure_enqueued(self) -> Tuple[bool, int]:
        """
        Ensure the job is indeed in its queue. See the docs for Queue.ensure_enqueued() for more information.
        """
        if self.has_started:
            raise AlreadyAcquired(
                "Job {id} has already been started, will not enqueue".format(id=self.id)
            )
        if self.has_finished:
            raise AlreadyResulted(
                "Job {id} has already been finished, will not enqueue".format(
                    id=self.id
                )
            )
        status = self.status
        if status in (JobStatus.SUCCESS, JobStatus.FAILED, JobStatus.CANCELLED):
            raise InvalidStatus(
                "Job {id} has status {status}, will not enqueue".format(
                    id=self.id, status=status
                )
            )
        return self.get_queue().ensure_enqueued(self)

    @property
    def redis_key(self) -> str:
        return "%s%s" % (JOB_KEY_PREFIX, self.id)

    @property
    def result_redis_key(self) -> str:
        return "%s%s" % (RESULT_KEY_PREFIX, self.id)

    @property
    def acquisition_info(self) -> Optional[dict]:
        # Acquisition info is always stored as JSON, not with the `encoding`
        acquisition_json = self.redis.hget(self.redis_key, "acquired")
        if acquisition_json:
            return json.loads(acquisition_json.decode())
        return None

    @property
    def has_finished(self) -> int:
        return self.redis.exists(self.result_redis_key)

    @property
    def has_started(self) -> bool:
        return self.redis.hexists(self.redis_key, "acquired")

    @property
    def encoded_result(self) -> Optional[bytes]:
        return self.redis.get(self.result_redis_key)

    @property
    def result(self) -> Optional[Any]:
        result_data = self.encoded_result
        if result_data is not None:
            return self.get_encoding().decode(result_data)
        return None

    @property
    def encoded_meta(self) -> Optional[bytes]:
        return self.redis.hget(self.redis_key, "meta")

    @property
    def meta(self) -> Optional[Any]:
        """
        Get any possible in-band progress metadata for this job.
        """
        meta_data = self.encoded_meta
        if meta_data is not None:
            return self.get_encoding().decode(meta_data)
        return None

    @property
    def status(self) -> JobStatus:
        return JobStatus(self.redis.hget(self.redis_key, "status").decode())  # type: ignore

    @property
    def heartbeat(self) -> Optional[float]:
        """
        Get heartbeat of this job after job's refresh_heartbeat has been called
        """
        heartbeat = self.redis.hget(self.redis_key, "heartbeat")
        if heartbeat is None:
            return None
        return float(heartbeat)  # type: ignore

    @property
    def exists(self) -> int:
        return self.redis.exists(self.redis_key)

    @property
    def result_ttl(self) -> int:
        return int(self.redis.hget(self.redis_key, "result_ttl"))  # type: ignore

    @property
    def duration(self) -> float:
        return float(self.redis.hget(self.redis_key, "duration"))  # type: ignore

    @property
    def queue_name(self) -> str:
        return self.redis.hget(self.redis_key, "queue").decode()  # type: ignore

    @property
    def kwargs(self) -> dict:
        kwargs_data = self.redis.hget(self.redis_key, "kwargs")
        if not kwargs_data:
            return {}
        return self.get_encoding().decode(kwargs_data)

    @property
    def callable_name(self) -> str:
        return self.redis.hget(self.redis_key, "callable").decode()  # type: ignore

    @property
    def encoding_name(self) -> str:
        encoding_name = self.redis.hget(self.redis_key, "encoding_name")
        if encoding_name:
            return encoding_name.decode()
        # If there is no encoding name, assume this job is from an earlier version,
        # and use the default encoding.
        return encoding.default_encoding_name or "no default encoding set"

    def get_encoding(self) -> encoding.BaseEncoding:
        return encoding.registry[self.encoding_name]()

    def get_queue(self) -> "Queue":
        from minique.models.queue import Queue

        return Queue(redis=self.redis, name=self.queue_name)

    def set_meta(self, meta: Any):
        """
        Set the "in-band" progress metadata for this job.
        Keep it short, though, for performance's sake...
        """
        self.redis.hset(self.redis_key, "meta", self.get_encoding().encode(meta))

    def refresh_heartbeat(self):
        """
        Set time.time() into heartbeat.
        """
        self.redis.hset(self.redis_key, "heartbeat", time.time())

    def __str__(self):
        return "<job %s>" % self.id

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Job) and (self.id == other.id)



import random
from contextlib import contextmanager
from importlib import import_module
from threading import local
from typing import Any, Callable, List, Optional

from minique.compat import TYPE_CHECKING

if TYPE_CHECKING:
    from minique.models.job import Job

_current_jobs = local()


def import_by_string(spec: str) -> Any:
    module_name, _, func_name = spec.rpartition(".")
    module = import_module(module_name)
    func = getattr(module, func_name)
    return func


@contextmanager
def _set_current_job(job: "Job"):
    assert not get_current_job()
    _current_jobs.current_job = job
    try:
        yield
    finally:
        _current_jobs.current_job = None


def get_current_job() -> Optional["Job"]:
    return getattr(_current_jobs, "current_job", None)


class cached_property(object):
    """A property that is only computed once per instance and then replaces
    itself with an ordinary attribute. Deleting the attribute resets the
    property.

    Source: https://github.com/bottlepy/bottle/blob/0.11.5/bottle.py#L175
    """

    def __init__(self, func: Callable) -> None:
        self.__doc__ = getattr(func, "__doc__", "")
        self.func = func

    def __get__(self, obj: Any, cls: Any) -> Any:
        if obj is None:
            # We're being accessed from the class itself, not from an object
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


consonants = "cdgkklmmnnprst"
vowels = "aeiou"


def get_random_pronounceable_string(length: int = 12) -> str:
    s = []  # type: List[str]
    while length > 0 and len(s) < length:
        s.append(random.choice(consonants))
        s.append(random.choice(vowels) * random.choice((1, 1, 2)))
    return "".join(s)[:length]



try:
    from typing import TYPE_CHECKING  # "New in version 3.5.2."
except ImportError:  # pragma: no cover
    TYPE_CHECKING = False

try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None  # type: ignore

__all__ = [
    "sentry_sdk",
    "TYPE_CHECKING",
]
